import java.text.DecimalFormat;
import java.util.ArrayList;

public final class SpaceManager
{
	ArrayList<Ship>     ShipList;
	ArrayList<Bullet>   BulletList;
	ArrayList<Planet>   PlanetList;

	ArrayList<IMShip>   ImmutableShipArray;
	ArrayList<IMBullet> ImmutableBulletArray;
	ArrayList<IMPlanet> ImmutablePlanetArray;
	
	private long       Frames;

	public SpaceManager()
	{
		ShipList   = new ArrayList<Ship>();
		BulletList = new ArrayList<Bullet>();
		PlanetList = new ArrayList<Planet>();
	
		Frames = 0;
	}

	public void UpdateIMShipArray ()
	{
		ImmutableShipArray   = new ArrayList<IMShip>();
		ImmutableShipArray.ensureCapacity(ShipList.size());
	
		for (Ship CurrShip : ShipList)
			ImmutableShipArray.add(CurrShip.GetIMShip());
	}

	public void UpdateIMBulletArray ()
	{
		ImmutableBulletArray   = new ArrayList<IMBullet>();
		ImmutableBulletArray.ensureCapacity(BulletList.size());
	
		for (Bullet CurrBullet : BulletList)
			ImmutableBulletArray.add(CurrBullet.GetIMBullet());
	}

	public void UpdateIMPlanetArray ()
	{
		ImmutablePlanetArray   = new ArrayList<IMPlanet>();
		ImmutablePlanetArray.ensureCapacity(PlanetList.size());
	
		for (Planet CurrPlanet : PlanetList)
			ImmutablePlanetArray.add(CurrPlanet.GetIMPlanet());
	}

	public ArrayList<Bullet> GetBulletList ()
	{
		return BulletList;
	}

	public ArrayList<Ship> GetShipList ()
	{
		return ShipList;
	}

	public ArrayList<Planet> GetPlanetList()
	{
		return PlanetList;
	}

	private void DoBullets ()
	{
		ArrayList<Bullet> DeadBullets = new ArrayList<Bullet>();
		ArrayList<Ship>   DeadShips = new ArrayList<Ship>();
	
		for (Bullet CurrBullet : BulletList)
		{
			Vector CurrBulletPos = CurrBullet.GetPosVector();
			Ship   BulletOwner   = CurrBullet.GetOwner();

			float Bx = CurrBulletPos.GetX();
			float By = CurrBulletPos.GetY();

			if ((Bx < 0) || (Bx > Constants.PLAY_FIELD_X_SIZE) ||
			    (By < 0) || (By > Constants.PLAY_FIELD_Y_SIZE))
			{
				BulletOwner.SetMisses(CurrBullet.GetOwner().GetMisses() + 1);
				BulletOwner.GetTeam().SetMisses(BulletOwner.GetTeam().GetMisses() + 1);
				DeadBullets.add(CurrBullet);
				continue;
			}

			for (Ship CurrShip : ShipList)
			{
				Vector CurrShipPos = CurrShip.GetPosVector();

				if (!(Constants.ALLOW_FRIENDLY_FIRE) &&
				   (BulletOwner.GetTeam().GetID() == CurrShip.GetTeam().GetID()))
					continue;
				
				if (Collision.HasCollided(CurrShip, CurrBullet))
				{
					int CurrShipHealth = CurrShip.GetHealth();

					if (CurrShipHealth == 1)
					{
						BulletOwner.SetKills(BulletOwner.GetKills() + 1);
						BulletOwner.GetTeam().SetKills(BulletOwner.GetTeam().GetKills() + 1);
						DeadShips.add(CurrShip);
					}
					else
					{
						CurrShip.SetHealth(CurrShipHealth - 1);
					}

					BulletOwner.SetHits(BulletOwner.GetHits() + 1);
					BulletOwner.GetTeam().SetHits(BulletOwner.GetTeam().GetHits() + 1);
					DeadBullets.add(CurrBullet);
				}
			}

			if (CurrBullet.GetTTL() == 0)
			{
				CurrBullet.GetOwner().SetMisses(CurrBullet.GetOwner().GetMisses() + 1);
				BulletOwner.GetTeam().SetMisses(BulletOwner.GetTeam().GetMisses() + 1);
				DeadBullets.add(CurrBullet);
			}
			else
			{
				CurrBullet.SetTTL(CurrBullet.GetTTL() - 1);

				if (Constants.GRAVITY_AFFECTS_BULLETS)
				{
					for (Planet CurrPlanet : PlanetList)
					{
						CurrBullet.GetVelVector().Add(
						   CurrPlanet.GetGravityAcceleration(CurrBullet));
					}
				}

				CurrBullet.GetPosVector().Add(CurrBullet.GetVelVector());
			}

			for (Planet CurrPlanet : PlanetList)
			{
				if (Collision.HasCollided(CurrPlanet, CurrBullet))
				{
					BulletOwner.SetMisses(CurrBullet.GetOwner().GetMisses() + 1);
					BulletOwner.GetTeam().SetMisses(BulletOwner.GetTeam().GetMisses() + 1);
					DeadBullets.add(CurrBullet);
				}
			}
		}
		
		BulletList.removeAll(DeadBullets);
		ShipList.removeAll(DeadShips);
	}

	private void DoAI ()
	{
		ArrayList<Ship> DeadShips = new ArrayList<Ship>();	
			
		UpdateIMBulletArray();
		UpdateIMShipArray();
		UpdateIMPlanetArray();

		for (Ship CurrShip : ShipList)
		{
			AI     CurrShipAI       = CurrShip.GetAI();
			Vector CurrShipAcc      = CurrShip.GetAccVector();
			double CurrShipHeading  = CurrShip.GetHeading();

			CurrShipAI.RunAI(ImmutableShipArray, ImmutableBulletArray,
			                 ImmutablePlanetArray, CurrShip.GetIMShipExt());

			if (CurrShipAI.GetTurningPercentage() != 0)
			{
				float NewHeading = (float)(CurrShip.GetHeading() +
				                          ((CurrShipAI.GetTurningPercentage() / 100.0) *
				                          Constants.SHIP_TURNING_RADS));

				CurrShip.SetHeading(Angle.WrapAngle(NewHeading));
			}

			Vector Heading = new Vector((float)Math.cos(CurrShipHeading), (float)Math.sin(CurrShipHeading));
			float  FonM    = (float)(CurrShip.GetThrust() *
			                         (CurrShipAI.GetThrustPercentage() / 100.0))
			                         / CurrShip.GetMass();
				
			CurrShipAcc.Set(Vector.Multiply(Heading, FonM));

			Vector CurrShipVel = CurrShip.GetVelVector();
			Vector NewVelocity = Vector.Add(CurrShipVel, CurrShipAcc);

			if (NewVelocity.GetLength() < Constants.SHIP_MAX_VEL)
				CurrShipVel.Set(NewVelocity);

			Vector GravVel = new Vector(0, 0);

			for (Planet CurrPlanet : PlanetList)
			{
				GravVel.Add(CurrPlanet.GetGravityAcceleration(CurrShip));
			}
			
			Vector CurrShipPos = CurrShip.GetPosVector();

			CurrShipPos.Add(CurrShipVel);
			CurrShipPos.Add(GravVel);
			
			if (CurrShipPos.GetX() < Constants.SHIP_RADIUS)
				CurrShipPos.SetX(Constants.SHIP_RADIUS);
			else if (CurrShipPos.GetX() > (Constants.PLAY_FIELD_X_SIZE - Constants.SHIP_RADIUS))
				CurrShipPos.SetX(Constants.PLAY_FIELD_X_SIZE);

			if (CurrShipPos.GetY() < Constants.SHIP_RADIUS)
				CurrShipPos.SetY(Constants.SHIP_RADIUS);
			else if (CurrShipPos.GetY() > (Constants.PLAY_FIELD_Y_SIZE - Constants.SHIP_RADIUS))
				CurrShipPos.SetY(Constants.PLAY_FIELD_Y_SIZE);

			for (Planet CurrPlanet : PlanetList)
			{
				Vector PlanetPos  = CurrPlanet.GetPosVector();

				if (Collision.HasCollided(CurrShip, CurrPlanet))
				{
					float  PlanetRadius      = (CurrPlanet.GetDiameter() / 2);
					float  ShipRadius        = (CurrShip.GetDiameter() / 2);
					Vector ShipPosFromPlanet;
					
					ShipPosFromPlanet = Vector.Normal(CurrPlanet.GetPosVector(), CurrShipPos);
					ShipPosFromPlanet.Unit();
					ShipPosFromPlanet.Multiply(PlanetRadius + ShipRadius + 1);
					ShipPosFromPlanet.Add(PlanetPos);
	
					CurrShipPos.Set(ShipPosFromPlanet);

					Vector[] ShipVel = Collision.GetVelAfterCollision(CurrShip.GetVelVector(),
					                                                  CurrShip.GetPosVector(),
					                                                  CurrShip.GetMass(),
					                                                  new Vector(0, 0),
					                                                  CurrPlanet.GetPosVector(),
					                                                  CurrPlanet.GetMass());

					CurrShipVel.Set(ShipVel[0].Multiply(1.5f));
					
					if (Constants.DAMAGE_ON_PLANET_COLLISON)
					{
						int CurrShipHealth = CurrShip.GetHealth();

						if (CurrShipHealth == 1)
							DeadShips.add(CurrShip);
						else
							CurrShip.SetHealth(CurrShipHealth - 1);
					}

					break;
				}
			}

			if (CurrShip.GetTimeToFire() > 0)
			{
				CurrShip.SetTimeToFire(CurrShip.GetTimeToFire() - 1);				
			}
			else if (CurrShip.GetAI().GetFire() == true) 
			{
				CurrShip.SetTimeToFire(Constants.SHIP_FRAMES_BETWEEN_FIRES);

				Vector CurrBulletPos = Vector.Add(CurrShip.GetPosVector(),
				                                  Vector.Multiply(Heading, Constants.SHIP_DIAMETER));

				Vector CurrBulletVel = Vector.Multiply(Heading, Constants.BULLET_SPEED);
				
				BulletList.add(new Bullet(CurrBulletPos, CurrBulletVel, CurrShipHeading, CurrShip));
			}
		}
		
		ShipList.removeAll(DeadShips);
	}

	public void PrintPlayingShips(boolean ShowHitMissKill)
	{
		int  TotalShips  = 0;
		Team WinningTeam = null;

		System.out.print("\n\n======= Currently Playing Ships =======\n");

		for (Ship CurrShip : ShipList)
		{
			StringBuilder ShipString = new StringBuilder("\n  \"");

			if (WinningTeam == null)
				WinningTeam = CurrShip.GetTeam();

			ShipString.append(CurrShip.GetName());
			ShipString.append("\"\n     -> Team \"");
			ShipString.append(CurrShip.GetTeam().GetName());
			ShipString.append("\", Health ");
			ShipString.append(CurrShip.GetHealth());
			ShipString.append("\n     -> Using AI \"");
			ShipString.append(CurrShip.GetAI().GetName());
			ShipString.append("\" made by ");
			ShipString.append(CurrShip.GetAI().GetCreator());

			if (ShowHitMissKill == true)
			{
				ShipString.append("\n     -> Hits: ");
				ShipString.append(CurrShip.GetHits());
				ShipString.append("   Misses: ");
				ShipString.append(CurrShip.GetMisses());
				ShipString.append("   Kills: ");
				ShipString.append(CurrShip.GetKills());

				int TotalShots = (CurrShip.GetHits() + CurrShip.GetMisses());
				
				if (TotalShots > 0)
				{
					DecimalFormat DF                 = new DecimalFormat("###.##");
					double        AccuracyPercentage = ((CurrShip.GetHits() * 100.0) / TotalShots);

					ShipString.append("\n     -> Accuracy: ");
					ShipString.append(DF.format(AccuracyPercentage));
					ShipString.append("%");
				}
			}
		 
  			System.out.println(ShipString.toString());
		}

		System.out.println("\n=======================================\n");
		System.out.println("  Total: " + ShipList.size());

		if (ShowHitMissKill == true)
		{
			System.out.println("\n=======================================\n");

			int           TotalShots         = (WinningTeam.GetHits() + WinningTeam.GetMisses());

			DecimalFormat DF                 = new DecimalFormat("###.##");
			double        AccuracyPercentage = ((WinningTeam.GetHits() * 100.0) / TotalShots);

			System.out.println("  Winning Team: \"" + WinningTeam.GetName() + "\"");
			System.out.println("     -> Hits: " + WinningTeam.GetHits() + "  Misses: " + WinningTeam.GetMisses() + "  Kills: " + WinningTeam.GetKills());
			System.out.println("     -> Accuracy: " + DF.format(AccuracyPercentage) + "%");
		}

		System.out.println("\n=======================================\n\n");
	}

	public void DoGameLoop ()
	{
		DoBullets();
		DoAI();
		Frames++;
	}

	public long GetElapsedFrames ()
	{
		return Frames;
	}

	public boolean IsGameOver()
	{
		Team FirstTeam = null;

		if (ShipList.size() <= 1)
			return true;

		for (Ship CurrShip : ShipList)
		{
			if (FirstTeam == null)
			{
				FirstTeam = CurrShip.GetTeam();
				continue;
			}

			if (CurrShip.GetTeam().GetID() != FirstTeam.GetID())
				return false;
		}

		return true;
	}
}
